/* refentry2man
 * This file is based on sample program elements.c found in sample
 * directory of James Clark's expat package.
 * Copyright (C) 2000 Rafa Kleger-Rudomin
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <stdio.h>
#include <unistd.h>
#include "convgen.h"

/* GLOBAL VARIABLES */

int multiple_files_option = 0;	/* if -m option given, find every refentry in input stream
				   and create proper man files */ 
int be_quiet_option = 0;
int debug_option = 0;
/* END OF GLOBAL VARIABLES */


static int depth = 0;		/* self explained */
static Handlers stack[MAXDEPTH]; /* stack of handler function pointers */
static XML_Parser parser;	/* the only parser used */
static int stdout_descriptor;	/* used to descriptor while parsing to temp buffer is performed */
static const XML_Char *expected_name; /* used by 'wait_' functions provided to ignore some input */


void set_output_to_temp_buffer(TempBuffer *buf) {
  /*
  size_t size;
  char *bp;
  size_t len;
  FILE *stream;
  stream = open_memstream(&bp, &len);
  */

  stdout_descriptor = fileno(stdout);
  stdout = open_memstream (&(buf->str), &(buf->len));
}
void unset_output_to_temp_buffer() {
  fflush(stdout);
  stdout = fdopen(stdout_descriptor, "w");
}




/* Wrapper handlers start here */

void start_tag_handler(void *userdata, const XML_Char *name, const XML_Char **attr)
{
  StartHandler handler;
  if ( ++depth >= MAXDEPTH ) { fprintf(stderr,"Depth too big\n"); exit(-1); }
  find_handlers(name, &stack[depth]);
  handler = stack[depth].starthandler;
  if(handler != NULL) (*handler)(name, attr);
  XML_SetCharacterDataHandler(parser,stack[depth].datahandler);
  
}

void end_tag_handler(void *userdata, const XML_Char *name)
{
  EndHandler handler;
  handler = stack[depth].endhandler;
  if(handler != NULL) (*handler)(name);
  if ( --depth < 0 ) { fprintf(stderr,"Depth < 0 \n"); exit(-1); };
  XML_SetCharacterDataHandler(parser,stack[depth].datahandler);
  
}

int unknown_encoding_handler (void *encodingHandlerData,
			      const XML_Char *name,
			      XML_Encoding *info){
  int i;
  if (strcmp(name,"iso-8859-2") == 0) {
    for(i=0; i<256; i++) info->map[i] = i;
    info->convert = NULL;
    info->data = NULL;
    info->release = NULL;
    return 1;
  }
  return 0;
}

/*
void default_handler(void *userData,
		     const XML_Char *s,
		     int len) {
  int i;
  fprintf(stderr,"Default handler: ");
  for(i=0; i<len; i++) fputc(*s++, stderr);
}
*/



void wait_for_start_tag_handler(void *userdata, const XML_Char *name, const XML_Char **attr) {
  if (strcmp(name,expected_name) == 0) {
    XML_SetElementHandler(parser, start_tag_handler, end_tag_handler);
    start_tag_handler(userdata, name, attr);
  }
}
void wait_for_start_tag(const XML_Char *name) {
  expected_name = name;
  XML_SetElementHandler(parser, wait_for_start_tag_handler, NULL);
}

void wait_for_end_tag_handler(void *userdata, const XML_Char *name) {
  if (strcmp(name,expected_name) == 0) {
    XML_SetElementHandler(parser, start_tag_handler, end_tag_handler);
    end_tag_handler(userdata, name);
  }
}
void wait_for_end_tag(const XML_Char *name) {
  expected_name = name;
  XML_SetElementHandler(parser, NULL, wait_for_end_tag_handler);
}

void main_loop(XML_Parser parser, FILE *input) {
  XML_Char buf[BUFSIZ];
  int done;
  size_t len;

  do {
    len = fread(buf, 1, sizeof(buf), input);
    done = len < sizeof(buf);
    if (!XML_Parse(parser, buf, len, done)) {
      fprintf(stderr,
	      "%s at line %d\n",
	      XML_ErrorString(XML_GetErrorCode(parser)),
	      XML_GetCurrentLineNumber(parser));
      exit(1);
    }
  } while (!done);
}

int external_entity_handler(XML_Parser parentparser,
			    const XML_Char *context,
			    const XML_Char *base,
			    const XML_Char *systemId,
			    const XML_Char *publicId) {
  FILE *f;
  //XML_Parser parser;
  fprintf(stderr,"External entity found: context='%s' base='%s' systemId='%s' publicId='%s'\n",
	  context, base, systemId, publicId);

  /* silently ignore entities that has PUBLIC id (mostly DTD's) */
  if (publicId != NULL) return 1;

  if ((f = fopen(systemId,"r")) != NULL) {
    parser = XML_ExternalEntityParserCreate(parentparser, context, NULL);

    main_loop(parser,f);
    
    fclose(f);
    XML_ParserFree(parser);
    parser = parentparser;
    return 1;
  }
  return 0;
}

void main_parse(FILE *input) {
  /*
    use global parser ( for wait_for_start_tag() )
    XML_Parser parser; 
   */
  
  parser = XML_ParserCreate(NULL);
    
  XML_SetUnknownEncodingHandler(parser, unknown_encoding_handler, NULL);
    
  XML_SetExternalEntityRefHandler(parser, external_entity_handler);
    
  XML_SetParamEntityParsing(parser, XML_PARAM_ENTITY_PARSING_ALWAYS);
  
  /* XML_SetDefaultHandlerExpand(parser, default_handler); */
    
  wait_for_start_tag("refentry");

  main_loop(parser, input);

  XML_ParserFree(parser);
}


int main(int argc, char * const argv[])
{
  //XML_Char buf[BUFSIZ];
  //int done;
  
  char opt;
  int i;
  FILE *input;
  
  while ((opt = getopt(argc, argv, "mqd")) != -1) {
    switch (opt) {
    case 'm': multiple_files_option = 1; break;
    case 'q': be_quiet_option = 1; break;
    case 'd': debug_option = 1; break;
    }      
  }

  /* open first file as */
  if (!multiple_files_option && ((argc - optind) > 1)) {
    fprintf(stderr, "unless -m given, only one file is accepted\n");
    exit(1);
  }
  
    if (argc - optind) {
		for(i=optind; i<argc; i++) {
			input = freopen(argv[i], "r", stdin); /* read files, not stdin */
			if (input == NULL) {
				fprintf(stderr, "can't open %s file.\n", argv[i]);
			} else {
				main_parse(input);
			}
		}
	} else {   
    	main_parse(stdin);
	}
	
  

  
  
  return 0;
}
